# frozen_string_literal: true

require 'json'
require 'yaml'
require 'date'

require 'puppet/util'
require 'puppet/util/execution'
require 'puppet/util/yaml'

module Puppet::Provider::Lxd
end

class Puppet::Provider::Lxd::ResourceBase < Puppet::Provider
  def self.lxc_resource_type
    raise NotImplementedError
  end

  def self.lxc_command
    raise NotImplementedError
  end

  def self.build_command(*args, type: nil)
    type ||= lxc_resource_type
    # This is a lie, there is no instances command, but make our life easier
    if type == 'instance'
      args
    else
      [type, *args]
    end
  end

  def self.lxc_remote_env_var
    raise NotImplementedError
  end

  def self.parse_resource(data)
    raise NotImplementedError
  end

  def self.set_resource_provider_props(props, remote_name: nil)
    props[:ensure] ||= :present
    props[:lxc_remote_name] = remote_name unless remote_name.nil?
    props[:provider] = name
    props
  end

  def self.mk_resource_methods(*attrs)
    attrs = [:provider, *attrs].uniq
    attrs.each do |attr|
      attr = attr.intern
      next if attr == :name

      define_method(attr) do
        @property_hash[attr]
      end

      define_method(attr.to_s + "=") do |val|
        @property_hash[attr] = val
        self.change_op ||= :update
      end
    end

    define_method("ensure") do
      @property_hash[:ensure] || :absent
    end

    define_method("ensure=") do |val|
      @property_hash[:ensure] = val
    end
  end

  def self.get_lxc_env(remote_name: nil)
    env = {}
    env[lxc_remote_env_var] = remote_name unless remote_name.nil?
    env
  end

  def self.run_lxc(*args, remote_name: nil, **options)
    base_env = get_lxc_env(remote_name: remote_name)
    env = base_env.merge(options.delete(:custom_environment) || {})

    options = {
      :failonfail         => true,
      :combine            => false,
      :custom_environment => env,
    }.merge(options)

    cmd = [lxc_command] + args
    debug("running lxc: #{cmd.join(' ')} options: #{options}")

    Puppet::Util::Execution.execute(cmd, options)
  end

  def self.lxc_list(*args, type: nil, remote_name: nil)
    args = build_command("list", "--format", "json", *args, type: type)
    output = run_lxc(*args, remote_name: remote_name)
    JSON.parse(output)
  end

  def self.lxc_show(name, *args, type: nil, remote_name: nil, parse: true)
    # We can't actually use show here because it doesn´t return all the information
    # Unfortunately, list doesn't use exact matching, so we also need to look through
    # the output to find what we need.
    args = build_command("list", "--format", "json", name, *args, type: type)

    output = run_lxc(*args, remote_name: remote_name, failonfail: false)
    result = JSON.parse(output)

    if result.is_a?(Hash)
      v = result[name]
      return nil if v.nil?

      if parse
        parse_resource(v)
      else
        v
      end
    else
      result.each do |entry|
        parsed_items = parse_resource(entry)

        found = parsed_items.find { |r|
          r[:name] == name
        }
        if found
          if parse
            return found
          else
            return entry
          end
        end
      end

      nil
    end
  end

  def self.lxc_create(name, *args, type: nil, remote_name: nil)
    args = build_command("create", name, *args, type: type)
    run_lxc(*args, remote_name: remote_name, combine: true)
  end

  def self.lxc_delete(name, *args, type: nil, remote_name: nil)
    args = build_command("delete", name, *args, type: type)
    run_lxc(*args, remote_name: remote_name, combine: true)
  end

  def self.with_tempfile(content, &block)
    Tempfile.open('lxc_resource') do |tmpfile|
      if content.is_a?(String)
        tmpfile.write(content)
      else
        tmpfile.write(YAML.dump(content))
      end
      tmpfile.flush

      yield tmpfile
    end
  end

  def self.lxc_put(name, data, *args, type: nil, remote_name: nil)
    args = build_command("edit", name, *args, type: type)

    with_tempfile(data) do |tmpfile|
      run_lxc(*args, stdinfile: tmpfile.path, remote_name: remote_name, combine: true)
    end
  end

  def self.instances(remote_name: 'local')
    begin
      listed = lxc_list(remote_name: remote_name)
    rescue Puppet::ExecutionFailure => e
      warn("Failed to list instances: #{e}")
      return []
    end

    objs = listed.map { |obj|
      parsed = parse_resource(obj)
      parsed
    }.flatten.uniq

    objs.map { |obj|
      set_resource_provider_props(obj, remote_name: remote_name)
      prov = new obj
      prov
    }
  end

  def self.prefetch(resources)
    # Get resources from all the remotes that have resources configured
    remotes = resources.map { |_, resource|
      p = resource.parameters
      if (remote_name = p[:lxc_remote_name])
        remote_name = remote_name.value
      end
      remote_name ||= 'local'
    }.uniq

    provs = remotes.map { |remote|
      instances(remote_name: remote)
    }.flatten

    provs.each do |prov|
      if (resource = resources[prov.name])
        resource.provider = prov
      end
    end
  end

  attr_reader :change_op

  def change_op=(value)
    raise Puppet::Error, "Cannot set change_op twice" if !@change_op.nil?
    @change_op = value
  end

  def initialize(*args)
    super(*args)

    @prev_state = @property_hash.dup
    @change_op = nil
  end

  def set_resource_properties
    resource.class.validproperties.each do |property|
      @property_hash[property] = resource.should(property)
    end
  end

  def create
    self.change_op = :create
    set_resource_properties
  end

  def exists?
    !@prev_state[:ensure].nil? && @prev_state[:ensure] != :absent
  end

  def destroy
    self.change_op = :destroy
  end

  def refresh
    self.change_op = :update
  end

  def create_resource
    raise NotImplementedError
  end

  def delete_resource
    self.class.lxc_delete(
      resource.name, remote_name: @prev_state[:lxc_remote_name])
  end

  def put_resource
    raise NotImplementedError
  end

  def reload_state
    new_data = self.class.lxc_show(name, remote_name: self.lxc_remote_name)
    if new_data.nil?
      @property_hash.clear
    else
      self.class.set_resource_provider_props(new_data)
      @property_hash = new_data
    end
  end

  def properties_changed?(props = nil)
    props ||= self.class.resource_type.validproperties

    props.any? { |n|
      if n == :ensure || !@property_hash.key?(n)
        false
      else
        current = @prev_state[n]
        current = current.to_s if current.is_a?(Symbol)
        desired = @property_hash[n]
        desired = desired.to_s if desired.is_a?(Symbol)

        c = current != desired
        c
      end
    }
  end

  def current_config
    remote_name = @prev_state[:lxc_remote_name]

    cur_config = self.class.lxc_show(name, remote_name: remote_name, parse: false)
    cur_config
  end

  def needs_replacement?
    properties_changed?([:lxc_remote_name])
  end

  def flush
    op = @change_op
    return if op.nil?
    @change_op = nil

    if op == :update && needs_replacement?
      info("Resource #{self.class.resource_type.name}[#{resource.name}] requires replacement")
      op = :replace
    end

    case op
    when :create
      set_resource_properties
      create_resource
    when :replace
      delete_resource
      set_resource_properties
      create_resource
    when :update
      put_resource
    when :destroy
      delete_resource
    else
      raise Puppet::Error, "Unexpected operation"
    end

    reload_state
    @prev_state = @property_hash.dup
  end
end

module Puppet::Provider::Lxd
  def self.extended(base)
    base.send :commands, :lxc => "lxc"
  end

  def lxc_command
    command('lxc')
  end

  def lxc_remote_env_var
    'LXC_REMOTE'
  end
end

module Puppet::Provider::Incus
  def self.extended(base)
    base.send :commands, :incus => "incus"
    base.send :defaultfor, :incus_path => /.+/
  end

  def lxc_command
    command('incus')
  end

  def lxc_remote_env_var
    'INCUS_REMOTE'
  end
end
