# frozen_string_literal: true

require 'puppet/provider/lxd'

module Puppet::Provider::Lxd::Image
class Base < Puppet::Provider::Lxd::ResourceBase
  desc 'Manage profiles in LXD'

  def self.lxc_resource_type
    'image'
  end

  def self.remote_provider
    @remote_provider ||= Puppet::Type.type(:lxd_remote).provider(name)
    @remote_provider
  end

  def self.update_source_to_remote_name(update_source)
    return nil if update_source.nil?

    # Convert the update_source we get from LXC into a remote name.
    # Otherwise, when we create images using the `remote:alias` syntax, the resource
    # will be permanently out of sync, as we will not get the same syntax back
    remote = remote_provider.remotes.find { |r|
      r.address == update_source['server'] && r.protocol == update_source['protocol']
    }

    return nil if remote.nil?
    remote.name
  end

  def self.parse_update_source(update_source)
    return nil if update_source.nil?

    remote_name = update_source_to_remote_name(update_source)
    remote_name ||= update_source['address']
    name = update_source['alias']

    "#{remote_name}:#{name}"
  end

  def self.parse_resource(data)
    names = data['aliases'].map { |a| a['name'] }
    names.append(data['fingerprint'])

    names.map { |name|
      {
        :name          => name,
        :type          => data['type'].to_sym,
        :properties    => data['properties'],
        :fingerprint   => data['fingerprint'],
        :architecture  => data['architecture'],
        :auto_update   => data['auto_update'],
        :remote_source => parse_update_source(data['update_source']),
      }.reject {|_, v| v.nil? }
    }
  end

  def self.copy_image(
      src:, dst:, type:, copy_aliases: false, auto_update: false,
      remote_name: 'local')
    args = ['image', "copy", src, "#{remote_name}:", "--quiet"]
    args += ['--copy-aliases'] if copy_aliases
    args += ['--auto-update'] if auto_update
    args += ['--vm'] if type == :'virtual-machine'
    args += ['--alias', dst]

    run_lxc(*args, remote_name: remote_name, combine: true)
  end

  def self.set_image_prop(name, key, value, remote_name: nil)
    args = ['image', 'set-property', name, key, value]
    run_lxc(*args, remote_name: remote_name, combine: true)
  end

  mk_resource_methods \
    :architecture,
    :auto_update,
    :copy_aliases,
    :fingerprint,
    :lxc_remote_name,
    :profiles,
    :properties,
    :public,
    :remote_source,
    :type

  def initialize(*args)
    super(*args)
    @latest_fingerprint = nil
  end

  def create_resource
    put_resource(force_copy: true)
  end

  def prepare_config(base = {})
    profiles = self.profiles && self.profiles.uniq.sort

    values = {
      'auto_update' => self.auto_update,
      'properties'  => self.properties,
      'public'      => self.public,
      'profiles'    => profiles,
    }
    values = values.reject { |_, v| v.nil? }

    base.merge(values)
  end

  def put_resource(force_copy: false)
    if force_copy
      remote_name = self.lxc_remote_name
      img = self.name
      # We grab directly from the resource here, as the existing image might not
      # be from the same or any remote source at all
      src = self.remote_source

      self.class.copy_image(
        src: src, dst: img,
        type: self.type,
        copy_aliases: self.copy_aliases,
        remote_name: remote_name)
    end

    config = prepare_config(current_config)
    self.class.lxc_put(self.name, config, remote_name: remote_name)
  end

  def needs_replacement?
    super || properties_changed?([:architecture, :type, :remote_source, :fingerprint])
  end

  def generated_image_properties
    {
      "puppet.resource" => self.name,
      "puppet.remote_source" => resource[:remote_source],
    }
  end

  def latest_fingerprint
    return @latest_fingerprint if @latest_fingerprint

    remote_source = resource[:remote_source]
    type = resource[:type]

    raise Puppet::Error, 'Missing remote_source' if remote_source.nil?

    image_remote, _, image_name = remote_source.rpartition(":")

    images = self.class.lxc_list(
      remote_source, remote_name: resource[:lxc_remote_name])

    image = images.find { |img|
      next false if img['type'] != type.to_s

      aliases = img['aliases'].map { |a| a['name'] }
      aliases.include?(image_name)
    }

    return nil if image.nil?
    @latest_fingerprint = image['fingerprint']
    @latest_fingerprint
  end
end
end
