# frozen_string_literal: true

require 'puppet/provider/lxd'

module Puppet::Provider::Lxd::Instance
class Base < Puppet::Provider::Lxd::ResourceBase
  desc 'Manage Instances in LXD'

  def self.lxc_resource_type
    'instance'
  end

  def self.parse_resource(data)
    all_props = data['config']
    status = if data['state']['status'] == 'Running'
      :running
    else
      :stopped
    end

    image = all_props['user.puppet.launch-image'] || all_props['volatile.base_image']
    instance_type = all_props['user.puppet.instance-type']

    [{
      :name          => data['name'],
      :ensure        => :present,
      :architecture  => data['architecture'],
      :config        => all_props,
      :description   => data['description'],
      :devices       => data['devices'],
      :ephemeral     => data['ephemeral'],
      :image         => image,
      :instance_type => instance_type,
      :profiles      => data['profiles'].sort,
      :project       => data['project'],
      :state         => data['state'],
      :type          => data['type'].to_sym,
      :uuid          => all_props['volatile.uuid'],
      :status        => status,
    }]
  end

  def self.replacement_required_props
    super + [:architecture, :ephemeral, :type]
  end

  mk_resource_methods \
    :architecture,
    :config,
    :description,
    :devices,
    :ephemeral,
    :image,
    :instance_type,
    :lxc_remote_name,
    :profiles,
    :project,
    :state,
    :status,
    :type,
    :uuid

  def initialize(*args)
    super(*args)

    @status = nil
  end

  def generated_instance_configs
    # Intentionally use resource instead of self methods here, as the properties
    # might not yet be set depending on sync order
    {
      'user.puppet.resource' => name,
      'user.puppet.instance-type' => resource[:instance_type],
      'user.puppet.launch-image' => resource[:image],
    }.reject{ |_, v| v.nil? || v == :absent }
  end

  def prepare_config(base = {})
    profiles = self.profiles && self.profiles.uniq.sort

    values = {
      'name'        => resource.name,
      'config'      => self.config,
      'description' => self.description,
      'devices'     => self.devices,
      'ephemeral'   => self.ephemeral,
      'profiles'    => profiles,
      'project'     => self.project,
      'type'        => self.type.to_s,
    }.reject { |_, v| v.nil? }

    base.merge(values)
  end

  def run_status_command(status)
    remote_name = self.lxc_remote_name
    self.class.run_lxc(status, self.name, remote_name: remote_name, combine: true)
  end

  def apply_status(status = nil)
    status ||= self.status

    if @prev_state[:status] != :running
      if [:running, :restarted].include?(status)
        run_status_command('start')
      end
    elsif status == :stopped
      run_status_command('stop')
    elsif status == :restarted
      run_status_command('restart')
    end
  end

  def create_resource
    remote_name = self.lxc_remote_name
    args = ['create', self.image, self.name, "--quiet"]
    args += ['--vm'] if self.type == :'virtual-machine'
    args += ['--type', self.instance_type] if self.instance_type

    self.class.with_tempfile(prepare_config) do |tmpfile|
      self.class.run_lxc(
        *args, remote_name: remote_name, stdinfile: tmpfile.path, combine: true)
    end

    apply_status
  end

  def delete_resource
    if self.ensure != :absent && !resource[:allow_delete]
      raise Puppet::Error, "Instance #{resource.name} has 'allow_delete' disabled, refusing to delete it"
    end

    apply_status(:stopped)
    if !self.ephemeral
      # Ephemeral instances are deleted when stopped
      super
    end
  end

  def put_resource
    # if we already determined we need replacement, we shouldnt bother checking
    # properties again
    changed = self.properties_changed?
    if changed
      remote_name = self.lxc_remote_name
      cur_config = current_config

      config = prepare_config(cur_config)
      self.class.lxc_put(name, config, type: "config", remote_name: remote_name)
    end

    apply_status
  end

  def needs_replacement?
    super || properties_changed?([:architecture, :ephemeral, :image, :instance_type, :type])
  end

  def running?
    @prev_state[:status] == :running
  end
end
end
