# frozen_string_literal: true

require 'puppet/provider/lxd'

module Puppet::Provider::Lxd::Remote
class Base < Puppet::Provider::Lxd::ResourceBase
  desc 'Manage remotes in LXD'

  def self.lxc_resource_type
    'remote'
  end

  # Override as we dont group by remote name (we are the remotes!)
  def self.instances
    begin
      listed = lxc_list()
    rescue Puppet::ExecutionFailure => e
      warn("Failed to list instances: #{e}")
      return []
    end

    objs = listed.map { |obj| parse_resource(obj) }.flatten
    objs.map { |obj|
      set_resource_provider_props(obj)
      prov = new obj
      prov
    }
  end

  def self.prefetch(resources)
    provs = instances
    provs.each do |prov|
      if (resource = resources[prov.name])
        resource.provider = prov
      end
    end
  end

  def self.remotes
    # Cached set of remotes for other providers to use
    @remotes ||= instances
    @remotes
  end

  def self.invalidate_remotes!
    @remotes = nil
  end

  def self.parse_resource(data)
    name, item = data

    project = item['Project'].to_s
    project = nil if project.empty?

    auth_type = item['AuthType'].to_s
    auth_type = nil if auth_type.empty?

    [{
      :name      => name,
      :address   => item['Addr'],
      :auth_type => auth_type,
      :project   => project,
      :protocol  => item['Protocol'],
      :public    => item['Public'],
    }]
  end

  mk_resource_methods \
    :address,
    :auth_type,
    :project,
    :protocol,
    :public

  def create_resource
    args = [
      '--protocol', self.protocol
    ]
    if self.auth_type
      args += ['--auth-type', self.auth_type]
    end
    if self.project
      args += ['--project': self.project]
    end
    if self.public == true
      args += ['--public']
    end

    self.class.run_lxc(
      'remote', 'add', *args, self.name, self.address, combine: true)
    self.class.invalidate_remotes!
  end

  def put_resource
    delete_resource
    create_resource
  end

  def delete_resource
    self.class.run_lxc('remote', 'remove', resource.name, combine: true)
    self.class.invalidate_remotes!
  end

  def needs_replacement
    properties_changed?([
      :address, :auth_type, :project, :protocol, :public
    ])
  end
end
end
