# frozen_string_literal: true

require 'puppet/property/boolean'
require 'puppet/property/lxd/keyvalue'

Puppet::Type.newtype(:lxd_image) do
  @doc = 'Manage LXC profile resource'

  ensurable do
    self.defaultvalues

    newvalue(:latest) do
      if !provider.exists?
        provider.create
      else
        provider.fingerprint = provider.latest_fingerprint
      end
    end

    def insync?(is)
      case should
      when :present
        return is != :absent
      when :latest
        # Short-circuit packages that are not present
        return false if is == :absent

        latest = provider.latest_fingerprint
        current = provider.fingerprint
        return current == latest
      when :absent
        return is == :absent
      else
        raise Puppet::Error, "Unexpected ensure value #{should}"
      end
    end

    def retrieve
      provider.ensure
    end
  end

  newparam(:name, isnamevar: true) do
    desc 'Name of the resource'
  end


  newproperty(:type, namevar: true) do
    newvalues(:container, :'virtual-machine')
    desc 'Image type'
    defaultto { 'container' }
  end

  newproperty(:fingerprint) do
    desc 'Image fingerprint'

    def should
      raise ArgumentError, "fingerprint is read-only"
    end
  end

  newproperty(:profiles, array_matching: :all) do
    desc 'Profiles to include on instance'
  end

  newproperty(:architecture, readonly: true) do
    desc 'Image architecture'
  end

  newproperty(:public, parent: Puppet::Property::Boolean) do
    desc 'Make image public'
    defaultto { false }
  end

  newproperty(:auto_update, parent: Puppet::Property::Boolean) do
    desc 'Auto update enabled'
    defaultto { false }
  end

  newproperty(:remote_source) do
    desc 'Image remote source'

    validate do |value|
      image_remote, _, image_name = value.rpartition(":")
      if image_remote.empty? or image_name.empty?
        raise ArgumentError, "remote_source must be in <remote>:<image_name> format"
      end
    end
  end

  newparam(:copy_aliases, parent: Puppet::Parameter::Boolean) do
    desc 'Copy aliases from remote source when downloading'
    defaultto { false }
  end

  newproperty(:lxc_remote_name) do
    desc 'Name of LXD remote to connect to, omit for local'
    defaultto { 'local' }
  end

  # Keep last, as it might depend on other values
  newproperty(:properties, parent: Puppet::Property::Lxd::KeyValue) do
    desc 'Image properties'

    def extra_props
      provider.generated_image_properties
    end
  end

  autorequire(:lxd_remote) do
    remote_source_remote = self[:remote_source].split(":").first

    [
      remote_source_remote,
      (self[:lxc_remote_name] if self[:lxc_remote_name] != 'local'),
    ].compact
  end

  autorequire(:lxd_profile) do
    profiles = self[:profiles] || []
    profiles.uniq
  end

  def refresh
    provider.fingerprint = :latest
  end
end
