# frozen_string_literal: true

require 'puppet/property/boolean'
require 'puppet/property/lxd/keyvalue'

Puppet::Type.newtype(:lxd_instance) do
  @doc = 'Manage LXC instance resource'

  ensurable do
    self.defaultvalues

    newvalue(:running, event: :instance_started, invalidate_refreshes: true) do
      provider.create if !provider.exists?
      provider.status = :running
    end

    newvalue(:stopped, event: :instance_stopped) do
      provider.create if !provider.exists?
      provider.status = :stopped
    end

    def insync?(is)
    case should
      when :stopped
        return is != :absent && !provider.running?
      when :running
        return is != :absent && provider.running?
      else
        return is == should
      end
    end

    def retrieve
      provider.ensure || :absent
    end
  end

  newparam(:name, isnamevar: true) do
    desc 'Name of the resource'

    validate do |value|
      unless value =~ /^([[:alnum:]]|-)+$/
        raise ArgumentError, "Name must contain only alphanumeric chars and dashes"
      end
    end
  end

  newproperty(:architecture) do
    desc 'Architecture'
  end

  # Note: image and instance_type must come before config, as they need to be set
  # when the config is generated as we add custom properties based on them.

  newproperty(:image) do
    desc 'Image to launch instance'
  end

  newproperty(:instance_type) do
    desc 'Type (size) of instance, see LXD docs for reference'
  end

  newproperty(:description) do
    desc 'Description'
  end

  newproperty(:devices) do
    desc 'Devices to assign'
  end
  newproperty(:ephemeral, parent: Puppet::Property::Boolean) do
    desc 'Make instance ephemeral'
    defaultto { false }
  end

  newproperty(:profiles, array_matching: :all) do
    desc 'Profiles to include on instance'
  end

  newproperty(:project) do
    desc 'Project to create resource in'
    defaultto { 'default' }
  end

  newproperty(:type) do
    desc 'Type to launch instance, container or virtual machine'

    newvalues(:container, :'virtual-machine')
    defaultto :container
  end

  newproperty(:uuid) do
    desc 'UUID of existing instance'

    def should
      raise ArgumentError, "uuid is read-only"
    end
  end

  newparam(:restart_on_notify, parent: Puppet::Property::Boolean) do
    desc 'Restart instance when resource is refreshed'
    defaultto { true }
  end

  newparam(:allow_delete, parent: Puppet::Property::Boolean) do
    desc 'Replace or delete instance when required, otherwise operation will be refused'
    defaultto { false }
  end

  newproperty(:lxc_remote_name) do
    desc 'Name of LXD remote to connect to, omit for local'
    defaultto { 'local' }
  end

  # Keep last, as it might depend on other values
  newproperty(:config, parent: Puppet::Property::Lxd::KeyValue) do
    desc 'Config to use for instance'

    def extra_props
      provider.generated_instance_configs
    end
  end

  autorequire(:lxd_remote) do
    lxc_remote_name = self[:lxc_remote_name]
    image = self[:image]

    image_remote, _, image_name = image.rpartition(":")

    r = [
      (lxc_remote_name if lxc_remote_name && lxc_remote_name != 'local'),
      (image_remote if !image_remote.empty? && !image_remote.start_with?("https"))
    ].compact
    r
  end

  autorequire(:lxd_image) do
    image = self[:image]
    image_remote, _, image_name = image.rpartition(":")

    if image_remote.empty?
      r = [image]
      r
    else
      []
    end
  end

  autorequire(:lxd_profile) do
    profiles = self[:profiles] || []
    profiles.uniq
  end

  def refresh
    if self[:ensure] == :running && self[:restart_on_notify] == true
      provider.status = :restarted
    end
  end
end
